# Import the necessary workshop packages
import pandas as pd
import matplotlib.pyplot as plt
import glob
import numpy as np
from scipy import signal
import os

########################################################################################################################

# Apply a low pass Butterworth filter to the collected data.
def filterdata(sensordata, fs=200, fc=4):
    # fs = sampling frequency
    # fc = cut-off frequency

    sensordata_filt = np.zeros(sensordata.shape)
    w = fc / (fs / 2) # Normalise the frequency
    b, a = signal.butter(fc, w, "low")

    # filtfilt() filters the data in both directions, forward and backward
    sensordata_filt[:,0] = signal.filtfilt(b, a, sensordata[:,0])
    sensordata_filt[:,1] = signal.filtfilt(b, a, sensordata[:,1])
    sensordata_filt[:,2] = signal.filtfilt(b, a, sensordata[:,2])

    return sensordata_filt

# Extract motion sequence from all data captured
def cut_data(acc_filt, gyr_filt, freq):
    # acc_filt = filtered acceleration signal
    # gyr_filt = filtered gyroscope signal
    # freq = measurement frequency

    # calculate the vector norm of the acceleration
    acc_abs = np.linalg.norm(acc_filt, axis=1)

    # TODO check different values for height, and range of detected peaks
    # find peaks in the signal
    peaks,_ = signal.find_peaks(acc_abs, height=8, distance=freq/2)
    diff_peaks = np.diff(peaks)

    # find first and last part of motion sequence without motion
    gap1 = np.argmax(diff_peaks[:round(len(diff_peaks)/2.5)])
    gap2 = np.argmax(diff_peaks[-round(len(diff_peaks)/2.5):])
    gap2 = int(gap2 + np.shape(diff_peaks) - round(len(diff_peaks)/2.5))

    # cut sequences
    acc_cut = acc_filt[peaks[gap1+1]:peaks[gap2],:]
    gyr_cut = gyr_filt[peaks[gap1+1]:peaks[gap2],:]

    return acc_cut, gyr_cut

# Data collected with different smartphones might have different orientations due to different alignments of the sensor
# inside the phone. Additionally, the orientation of the smartphone inside a pocket might have been different.
# Principal component analysis (PCA) can be used to correct these errors by sorting the components dependent on their
# energy.
def rotate(X, theta, axis="x"):
    # Rotate multidimensional array `X` `theta` degrees around axis `axis`
    c, s = np.cos(np.deg2rad(theta)), np.sin(np.deg2rad(theta))
    if axis == "x": return np.dot(X, np.array([
        [1.,  0,  0],
        [0 ,  c, -s],
        [0 ,  s,  c]
        ]))
    elif axis == "y": return np.dot(X, np.array([
        [c,  0,  -s],
        [0,  1,   0],
        [s,  0,   c]
        ]))
    elif axis == "z": return np.dot(X, np.array([
        [c, -s,  0 ],
        [s,  c,  0 ],
        [0,  0,  1.],
        ]))

def rotate_data(acc_cut, gyr_cut):
    # acc_cut = motion sequence of acceleration signal
    # gyr_cut = motion sequence of gyroscope signal

    # perform a singular value decomposition to get directions with most energy
    cov = np.matmul(acc_cut.T, acc_cut)
    u,s,v = np.linalg.svd(cov, compute_uv=True)

    # rotation is stored in u, all sensors are equally aligned to the global coordinate system
    acc_pca = np.dot(acc_cut,u)
    gyr_pca = np.dot(gyr_cut,u)

    if np.mean(acc_pca[:,0]) < 0:
        acc_pca = rotate(acc_pca, 180, axis="z")
        gyr_pca = rotate(gyr_pca, 180, axis="z")

    return acc_pca, gyr_pca

# Exclude data if their RMSE to the mean exceeds a defined threshold
def rmse(predictions, targets):
    return np.sqrt(((predictions - targets) ** 2).mean())

def clean_sequences(IMU, threshold):
    # IMU_sequences = rotated and merged acceleration and gyroscope signal

    IMU_sequences = IMU[:,:-1,:].astype(float)

    # calculate mean sequence for each signal
    mean_IMU = np.mean(IMU_sequences, axis=0)

    sequences_clean = []
    for seq, seq1 in zip(IMU_sequences, IMU):
        # calculate RMSE of each sequence to the mean
        RMSE1 = rmse(seq[:,0], mean_IMU[:,0])
        RMSE2 = rmse(seq[:,1], mean_IMU[:,1])
        RMSE3 = rmse(seq[:,2], mean_IMU[:,2])
        RMSE4 = rmse(seq[:,3], mean_IMU[:,3])
        RMSE5 = rmse(seq[:,4], mean_IMU[:,4])
        RMSE6 = rmse(seq[:,5], mean_IMU[:,5])
        RMSE = RMSE1 + RMSE2 + RMSE3 + RMSE4+ RMSE5 + RMSE6

        # delete sequences exceeding a predefined threshold
        if RMSE < threshold:
            sequences_clean.append(seq1)

    sequences_clean = np.transpose(np.dstack(sequences_clean),(2,0,1))

    return sequences_clean

########################################################################################################################
# plot data
plot = False

# Load smartphone files
path = os.path.dirname(__file__)
files1 = glob.glob(os.path.join(path, 'DATA\\Smartphone1\\*'))
files2 = glob.glob(os.path.join(path, 'DATA\\Smartphone2\\*'))
files = np.sort(files1 + files2)

for file in files:
    # get file name
    phone = file.split('\\')[-2]
    name = file.split('\\')[-1]
    print(phone, name)

    # read data from file
    sensors = glob.glob(file + "/*.csv")

    # get acceleration
    data1 = pd.read_csv(sensors[0], sep=",", header=None, keep_default_na=False)
    acc = data1.iloc[2:, 1:4].values.astype(float)

    # get angular rate
    data2 = pd.read_csv(sensors[1], sep=",", header=None, keep_default_na=False)
    gyr = data2.iloc[2:, 1:4].values.astype(float)

    # get frequency (different smartphones use different measurement frequencies)
    timestamp = data2.iloc[1:, 0].values.astype(float)
    freq = int(round(1/(timestamp[1]-timestamp[0])))

    # filter data
    acc_filt = filterdata(acc, fs=freq)
    gyr_filt = filterdata(gyr, fs=freq)

    if plot:
        # plot to check filtering
        t = np.arange(acc.shape[0])/freq
        fig = plt.figure(figsize=(10,6), dpi=200)
        plt.subplot(1,2,1)
        plt.title("acceleration")
        plt.plot(t, acc[:,0], "--r", linewidth=0.5)
        plt.plot(t, acc_filt[:,0], "k", linewidth=0.5)
        plt.plot(t, acc[:,1], "--g", linewidth=0.5)
        plt.plot(t, acc_filt[:,1], "k", linewidth=0.5)
        plt.plot(t, acc[:,2], "--b", linewidth=0.5)
        plt.plot(t, acc_filt[:,2], "k", linewidth=0.5)
        plt.legend(["x", "x filt", "y", "y filt", "z", "z filt"])

        t = np.arange(gyr.shape[0]) / freq
        plt.subplot(1,2,2)
        plt.title("angular rate")
        plt.plot(t, gyr[:,0], "--r", linewidth=0.5)
        plt.plot(t, gyr_filt[:,0], "k", linewidth=0.5)
        plt.plot(t, gyr[:,1], "--g", linewidth=0.5)
        plt.plot(t, gyr_filt[:,1], "k", linewidth=0.5)
        plt.plot(t, gyr[:,2], "--b", linewidth=0.5)
        plt.plot(t, gyr_filt[:,2], "k", linewidth=0.5)
        plt.suptitle("Smartphone1 Filtered")
        plt.show()

    # cut data to motion sequence
    acc_cut, gyr_cut = cut_data(acc_filt, gyr_filt, freq)

    # plot to check if data is properly cut to motion sequence
    if plot:
        t = np.arange(acc_cut.shape[0])/freq
        fig = plt.figure(figsize=(10,6), dpi=200)
        plt.plot(t, acc_cut[:,0])
        plt.suptitle("Smartphone1 Cut")
        plt.show()

    # perform principal component analysis/singular value decomposition to align data
    # different smartphones have a different alignment of their sensors
    # smartphones placed loose in the backpockets result in different alignments
    acc_pca, gyr_pca = rotate_data(acc_cut, gyr_cut)

    # check if all data is aligned similarly
    if plot:
        t = np.arange(acc_pca.shape[0])/freq
        plt.figure(figsize=(10,6), dpi=200)

        plt.subplot(3,2,1)
        plt.plot(t, acc_pca[:,0], "-r", linewidth=0.5)
        plt.plot(t, acc_cut[:,0], ":r", linewidth=0.5)

        plt.subplot(3,2,3)
        plt.plot(t, acc_pca[:,1], "-g", linewidth=0.5)
        plt.plot(t, acc_cut[:,1], ":g", linewidth=0.5)

        plt.subplot(3,2,5)
        plt.plot(t, acc_pca[:,2], "-b", linewidth=0.5)
        plt.plot(t, acc_cut[:,2], ":b", linewidth=0.5)

        plt.subplot(3,2,2)
        plt.plot(t, gyr_pca[:,0], "-r", linewidth=0.5)
        plt.plot(t, gyr_cut[:,0], ":r", linewidth=0.5)

        plt.subplot(3,2,4)
        plt.plot(t, gyr_pca[:,1], "-g", linewidth=0.5)
        plt.plot(t, gyr_cut[:,1], ":g", linewidth=0.5)

        plt.subplot(3,2,6)
        plt.plot(t, gyr_pca[:,2], "-b", linewidth=0.5)
        plt.plot(t, gyr_cut[:,2], ":b", linewidth=0.5)

        plt.suptitle("Smartphone PCA")
        plt.show()

    # cut to similar sequences (strides, steps) based on y-axis angular rate
    # there a multiple solutions for which data to use and additional inputs to the function
    peaks, _ = signal.find_peaks(gyr_pca[:,2], distance=freq, height=1)

    # check if appropriate peaks are found in the data
    if plot:
        plt.figure(figsize=(10,6), dpi=200)
        plt.plot(gyr_pca[:,2])
        plt.plot(peaks, gyr_pca[peaks,2], "x")
        plt.suptitle("Smartphone Peaks")
        plt.show()

    # split sequences based on peaks found
    acc_sequences = np.split(acc_pca, peaks, axis=0)
    gyr_sequences = np.split(gyr_pca, peaks, axis=0)

    # resample sequences to 100 %
    acc_sequences_norm = [np.vstack(signal.resample(seq, 101)) for seq in acc_sequences]
    gyr_sequences_norm = [np.vstack(signal.resample(seq, 101)) for seq in gyr_sequences]

    # get 3D array sorted as NSF (n samples, s sequence, f features)
    acc_sequences_3D = np.transpose(np.dstack(acc_sequences_norm),(2,0,1))
    gyr_sequences_3D = np.transpose(np.dstack(gyr_sequences_norm),(2,0,1))

    # concatenate acceleration and angular rate to save as one array
    IMU_sequences = np.concatenate((acc_sequences_3D, gyr_sequences_3D), axis=2)

    # save data
    filename = phone + '_' + name
    np.save(os.path.join(path, 'DATA\\npy', filename), IMU_sequences)

########################################################################################################################
# Prepare inputs and outputs for classification

# Load smartphone files
path = os.path.dirname(__file__)
files = np.sort(glob.glob(os.path.join(path, 'DATA\\npy\\*.npy')))

# assign data structures
inputs = np.empty([0, 101, 6])
outputs = np.empty([0])
subjects = np.empty([0])
for file in files:
    # get file name
    name = file.split("\\")[-1]
    print(name)

    # get subject number and condition (normal, impaired)
    subject = str(np.array(name.split("_")[1].split("t")[1]))
    condition = name.split("_")[2].split(".")[0].split("0")[0]

    # load files
    data = np.load(file)
    num_samples = data.shape[0]

    # assign "0" and "1" as target values to condition
    if condition == "normal":
        output = 0
    else:
        output = 1

    # repeated copies of subject and output for each step
    output = (output,) * num_samples
    subject = (subject,) * num_samples

    # concatenate data of all subjects
    subjects = np.concatenate((subjects, subject), axis=0)
    outputs = np.concatenate((outputs, output), axis=0)
    inputs = np.concatenate((inputs, data), axis=0)

# split data according to condition for further processing
normal_idx = np.where(outputs == 0)[0]
normal_seq = inputs[normal_idx, :, :]
normal_subjects = subjects[normal_idx]

impaired_idx = np.where(outputs == 1)[0]
impaired_seq = inputs[impaired_idx, :, :]
impaired_subjects = subjects[impaired_idx]

# concatenate sequence and subject number
normal = np.concatenate([normal_seq, np.transpose(np.expand_dims(np.repeat
                        (np.expand_dims(normal_subjects, axis=1), 6, axis=1), axis=2), [0, 2, 1])], axis=1)
impaired = np.concatenate([impaired_seq, np.transpose(np.expand_dims(np.repeat
                          (np.expand_dims(impaired_subjects, axis=1), 6, axis=1), axis=2), [0, 2, 1])], axis=1)

# clean sequences based on threshold (6 to 14)
threshold = 14 # TODO try different threshold
normal_sequences = clean_sequences(normal, threshold).astype(float)
normal_outputs = np.zeros(normal_sequences.shape[0]).astype(float)
impaired_sequences = clean_sequences(impaired, threshold).astype(float)
impaired_outputs = np.ones(impaired_sequences.shape[0]).astype(float)

# store data in inputs and outputs
inputs_clean = np.concatenate([normal_sequences, impaired_sequences], axis=0)
outputs_clean = np.concatenate([normal_outputs, impaired_outputs], axis=0)

# save input data
np.save(os.path.join(path, 'DATA\\inputs_outputs', 'inputs'), inputs_clean)
np.save(os.path.join(path, 'DATA\\inputs_outputs', 'outputs'), outputs_clean)

# plot input data
if plot:
    fig = plt.figure(figsize=(10, 6), dpi=200)
    plt.subplot(3, 2, 1)
    plt.title("acceleration x")
    plt.plot(impaired_sequences[:, :-1, 0].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 0].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 0], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 0], axis=0), color="k")
    plt.xticks([])
    plt.ylabel('acc [m/s^2]')
    plt.subplot(3, 2, 3)
    plt.title("acceleration y")
    plt.plot(impaired_sequences[:, :-1, 1].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 1].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 1], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 1], axis=0), color="k")
    plt.xticks([])
    plt.ylabel('acc [m/s^2]')
    plt.subplot(3, 2, 5)
    plt.title("acceleration z")
    plt.plot(impaired_sequences[:, :-1, 2].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 2].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 2], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 2], axis=0), color="k")
    plt.ylabel('acc [m/s^2]')
    plt.xlabel('stride [%]')
    plt.subplot(3, 2, 2)
    plt.title("angular rate x")
    plt.plot(impaired_sequences[:, :-1, 3].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 3].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 3], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 3], axis=0), color="k")
    plt.ylabel('ang vel [rad/s]')
    plt.xticks([])
    plt.subplot(3, 2, 4)
    plt.title("angular rate y")
    plt.plot(impaired_sequences[:, :-1, 4].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 4].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 4], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 4], axis=0), color="k")
    plt.xticks([])
    plt.ylabel('ang vel [rad/s]')
    plt.subplot(3, 2, 6)
    plt.title("angular rate z")
    plt.plot(impaired_sequences[:, :-1, 5].transpose(), color="r")
    plt.plot(normal_sequences[:, :-1, 5].transpose(), color="b")
    plt.plot(np.mean(impaired_sequences[:, :-1, 5], axis=0), color="k")
    plt.plot(np.mean(normal_sequences[:, :-1, 5], axis=0), color="k")
    plt.ylabel('ang vel [rad/s]')
    plt.xlabel('stride [%]')
    plt.show()

