# Import the necessary workshop packages
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import seaborn as sn
import os
########################################################################################################################

# Load data
path = os.path.dirname(__file__)
inputs = np.load(os.path.join(path, 'DATA\\inputs_outputs\\inputs.npy'))
outputs = np.load(os.path.join(path, 'DATA\\inputs_outputs\\outputs.npy'))

# Dense Neural Network preparation
# INPUTS: flatten inputs and assign proper datatype to outputs
input_type = "angular rate" # "acceleration", "angular rate", "both"
if input_type == "acceleration":
    inputs_flat = inputs[:, :-1, :3].reshape(inputs[:, :-1, :3].shape[0], -1, order="C")
    inputs_flat = np.concatenate([inputs_flat, np.expand_dims(inputs[:, -1, 0], axis=1)], axis=1)
elif input_type == "angular rate":
    inputs_flat = inputs[:, :-1, 3:6].reshape(inputs[:, :-1, 3:].shape[0], -1, order="C")
    inputs_flat = np.concatenate([inputs_flat, np.expand_dims(inputs[:, -1, 0], axis=1)], axis=1)
else:
    inputs_flat = inputs[:, :-1, :].reshape(inputs[:, :-1, :].shape[0], -1, order="C")
    inputs_flat = np.concatenate([inputs_flat, np.expand_dims(inputs[:, -1, 0], axis=1)], axis=1)

# shuffle dataset
data_flat = np.concatenate([inputs_flat, np.expand_dims(outputs, axis=1)], axis=1)
np.random.shuffle(data_flat)

inputs_flat = data_flat[:, :-2]
outputs_flat = data_flat[:, -1]
subjects_flat = data_flat[:, -2]

################################################################################

# split data in test, validation and training based on a leave-one-subject-out validation
num_subjects = len(np.unique(subjects_flat))
confusion_matrix = []
for i in range(0, num_subjects):
    # define one subject's data as test data
    MLP_idx_test = np.squeeze(np.array(np.where(subjects_flat == i)))
    MLP_inputs_test = inputs_flat[MLP_idx_test, :]
    MLP_outputs_test = outputs_flat[MLP_idx_test]

    # use all left data as training and validation set
    MLP_idx_tmp = np.squeeze(np.array(np.where(subjects_flat != i)))
    MLP_inputs_tmp = inputs_flat[MLP_idx_tmp, :]
    MLP_outputs_tmp = outputs_flat[MLP_idx_tmp]

    # split validation set from training set
    MLP_train_size = int(0.8 * MLP_outputs_tmp.shape[0])
    MLP_inputs_train = MLP_inputs_tmp[:MLP_train_size, :]
    MLP_outputs_train = MLP_outputs_tmp[:MLP_train_size]
    MLP_inputs_val = MLP_inputs_tmp[MLP_train_size:, :]
    MLP_outputs_val = MLP_outputs_tmp[MLP_train_size:]

    # exclude participant with less than 10 steps
    if len(MLP_outputs_test) <= 10:
        continue

    # normalise training, validation, and test inputs
    train_mean = np.mean(np.mean(MLP_inputs_train, axis=0))
    train_std = np.mean(np.std(MLP_inputs_train, axis=0))

    MLP_inputs_train = (MLP_inputs_train - train_mean) / train_std
    MLP_inputs_val = (MLP_inputs_val - train_mean) / train_std
    MLP_inputs_test = (MLP_inputs_test - train_mean) / train_std

    # check data distribution to ensure the dataset is balanced
    num_pos_train = np.sum(MLP_outputs_train)
    num_pos_val = np.sum(MLP_outputs_val)
    num_pos_test = np.sum(MLP_outputs_test)

    # Define Dense Network
    MLP_model = tf.keras.models.Sequential([
        tf.keras.layers.Dense(128, activation="relu", kernel_regularizer=tf.keras.regularizers.l1_l2(l1=0, l2=0.01)),
        tf.keras.layers.Dropout(0.4),
        tf.keras.layers.Dense(1, activation="sigmoid")]) # sigmoid for output function values between 0 and 1

    # Compile model
    # set optimizer, loss and metric functions
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001)
    MLP_model.compile(optimizer=optimizer, loss="binary_crossentropy", metrics=["accuracy"])

    # Train and validate model
    history = MLP_model.fit(MLP_inputs_train, MLP_outputs_train, batch_size=64, epochs=50,
                            validation_data=(MLP_inputs_val, MLP_outputs_val), verbose=0)

    # Test model
    MLP_results = MLP_model.evaluate(MLP_inputs_test, MLP_outputs_test, batch_size=None, verbose=0)

    # Generate predictions (probabilities -- the output of the last layer)
    MLP_predictions = MLP_model.predict(MLP_inputs_test)
    MLP_predictions = np.round(MLP_predictions, decimals=0)

    # Create confusion matrix
    cm = tf.math.confusion_matrix(MLP_outputs_test, MLP_predictions,
                                  num_classes=2, weights=None, dtype=tf.dtypes.int32)

    confusion_matrix.append(cm.numpy())

    # fig = plt.figure()
    # sn.heatmap(cm, annot=True)
    # plt.xlabel("Predicted Class")
    # plt.ylabel("Target Class")
    # plt.title("MLP Confusion Matrix")
    # plt.show()

cm_overall = np.sum(np.dstack(confusion_matrix), axis=2)

fig = plt.figure()
sn.heatmap(cm_overall, annot=True)
plt.xlabel("Predicted Class")
plt.ylabel("Target Class")
plt.title("MLP Confusion Matrix")
plt.show()

accuracy = (cm_overall[0, 0]+cm_overall[1, 1])/(cm_overall[0, 0]+cm_overall[1, 1]+cm_overall[1, 0]+cm_overall[0, 1])
precision = cm_overall[0, 0]/(cm_overall[0, 0]+cm_overall[0, 1])
recall = cm_overall[0, 0]/(cm_overall[0, 0]+cm_overall[1, 0])
print("Accuracy:", accuracy, "Precision:", precision, "Recall:", recall)
